@@ -15,13 +15,14 @@ utility subroutines used by B<Astro::Coord::ECI> and its descendents.
 What ended up here was anything that was essentially a subroutine, not
 a method.
 
-This package exports nothing by default. But all the constants and
-subroutines documented below are exportable, and the :all tag gets you
-all of them.
+This package exports nothing by default. But all the constants,
+variables, and subroutines documented below are exportable, and the :all
+tag gets you all of them.
 
-The following constants are exportable:
+=head2 The following constants are exportable:
 
  AU = number of kilometers in an astronomical unit
+ JD_OF_EPOCH = the Julian Day of Perl epoch 0
  LIGHTYEAR = number of kilometers in a light year
  PARSEC = number of kilometers in a parsec
  PERL2000 = January 1 2000, 12 noon universal, in Perl time
@@ -30,7 +31,32 @@ The following constants are exportable:
  SECSPERDAY = the number of seconds in a day
  TWOPI = twice the circle ratio
 
-In addition, the following subroutines are exportable:
+=head2 The following global variable is exportable:
+
+=head3 $DATETIMEFORMAT
+
+This variable represents the POSIX::strftime format used to convert
+times to strings. The default value is '%a %b %d %Y %H:%M:%S' to be
+consistent with the behavior of gmtime (or, to be precise, the
+behavior of ctime as documented on my system).
+
+=head3 $JD_GREGORIAN
+
+This variable represents the Julian Day of the switch from Julian to
+Gregorian calendars. This is used by date2jd(), jd2date(), and the
+routines which depend on them, for deciding whether the date is to be
+interpreted as in the Julian or Gregorian calendar. Its initial setting
+is 2299160.5, which represents midnight October 15 1582 in the Gregorian
+calendar, which is the date that calendar was first adopted. This is
+slightly different than the value of 2299161 (noon of the same day) used
+by Jean Meeus.
+
+If you are interested in historical calculations, you may wish to reset
+this appropriately. If you use date2jd to calculate the new value, be
+aware of the effect the current setting of $JD_GREGORIAN has on the
+interpretation of the date you give.
+
+=head2 In addition, the following subroutines are exportable:
 
 =over 4
 
@@ -41,21 +67,23 @@ use warnings;
 
 package Astro::Coord::ECI::Utils;
 
-our $VERSION = '0.007';
+our $VERSION = '0.008';
 our @ISA = qw{Exporter};
 
 use Carp;
 use Data::Dumper;
-use POSIX qw{floor};
+use POSIX qw{floor strftime};
 use Time::Local;
 use UNIVERSAL qw{can isa};
 
 our @EXPORT;
 our @EXPORT_OK = qw{
-	AU LIGHTYEAR PARSEC PERL2000 PI PIOVER2 SECSPERDAY TWOPI acos
-	asin atmospheric_extinction deg2rad distsq dynamical_delta
-	equation_of_time find_first_true intensity_to_magnitude
-	jcent2000 jday2000 julianday mod2pi nutation_in_longitude
+	AU $DATETIMEFORMAT $JD_GREGORIAN JD_OF_EPOCH LIGHTYEAR PARSEC
+	PERL2000 PI PIOVER2 SECSPERDAY TWOPI acos asin
+	atmospheric_extinction date2epoch date2jd deg2rad distsq
+	dynamical_delta epoch2datetime equation_of_time find_first_true
+	intensity_to_magnitude jcent2000 jd2date jd2datetime jday2000
+	julianday load_module mod2pi nutation_in_longitude
 	nutation_in_obliquity obliquity omega rad2deg tan theta0
 	thetag};
 
@@ -105,8 +133,6 @@ value.
 sub asin {atan2 ($_[0], sqrt (1 - $_[0] * $_[0]))}
 
 
-=for comment help syntax-highlighting editor "
-
 =item $magnitude = atmospheric_extinction ($elevation, $height);
 
 This subroutine calculates the typical atmospheric extinction in
@@ -122,8 +148,6 @@ this article makes it clear that the actual value of the
 atmospheric extinction can vary greatly from the typical
 values given even in the absence of cloud cover.
 
-=for comment help syntax-highlighting editor "
-
 =cut
 
 #	Note that the "constant" 0.120 in Aaer (aerosol scattering) is
@@ -146,6 +170,85 @@ my $Aaer = 0.120 * exp (-$height / 1.5);	# Green 4
 }
 
 
+=item $jd = date2jd ($sec, $min, $hr, $day, $mon, $yr)
+
+This subroutine converts the given date to the corresponding Julian day.
+The inputs are as for B<Time::Local::timegm>; $mon is in the range 0 -
+11, and $yr is from 1900, with earlier years being negative. The year 1
+BC is represented as -1900.
+
+If less than 6 arguments are provided, zeroes will be prepended to the
+argument list as needed.
+
+The date is presumed to be in the Gregorian calendar. If the resultant
+Julian Day is before $JD_GREGORIAN, the date is reinterpreted as being
+from the Julian calendar.
+
+The only validation is that the month be between 0 and 11 inclusive, and
+that the year be not less than -6612 (4713 BC). Fractional days are
+accepted.
+
+The algorithm is from Jean Meeus' "Astronomical Algorithms", second
+edition, chapter 7 ("Julian Day"), pages 60ff, but the month is
+zero-based, not 1-based, and years are 1900-based.
+
+=cut
+
+our $DATETIMEFORMAT;
+our $JD_GREGORIAN;
+BEGIN {
+    $DATETIMEFORMAT = '%a %b %d %Y %H:%M:%S';
+    $JD_GREGORIAN = 2299160.5;
+}
+
+sub date2jd {
+    unshift @_, 0 while @_ < 6;
+    my ($sec, $min, $hr, $day, $mon, $yr) = @_;
+    $mon++;		# Algorithm expects month 1-12.
+    $yr += 1900;	# Algorithm expects zero-based year.
+    $yr < -4712 and croak "Error - Invalid year $yr";
+    $mon < 1 || $mon > 12 and croak "Error - Invalid month $mon";
+    if ($mon < 3) {
+	--$yr;
+	$mon += 12;
+    }
+    my $A = floor ($yr / 100);
+    my $B = 2 - $A + floor ($A / 4);
+    my $jd = floor (365.25 * ($yr + 4716)) +
+	floor (30.6001 * ($mon + 1)) + $day + $B - 1524.5 +
+	((($sec || 0) / 60 + ($min || 0)) / 60 + ($hr || 0)) / 24;
+    $jd < $JD_GREGORIAN and
+	$jd = floor (365.25 * ($yr + 4716)) +
+	floor (30.6001 * ($mon + 1)) + $day - 1524.5 +
+	((($sec || 0) / 60 + ($min || 0)) / 60 + ($hr || 0)) / 24;
+    $jd;
+}
+
+use constant JD_OF_EPOCH => date2jd (gmtime (0));
+
+
+=item $epoch = date2epoch ($sec, $min, $hr, $day, $mon, $yr)
+
+This is a convenience routine that converts the given date to seconds
+since the epoch, going through date2jd() to do so. The arguments are the
+same as those of date2jd().
+
+If less than 6 arguments are provided, zeroes will be prepended to the
+argument list as needed.
+
+The functionality is the same as B<Time::Local::timegm>, but this
+function lacks timegm's limited date range.
+
+=cut
+
+sub date2epoch {
+    unshift @_, 0 while @_ < 6;
+    my ($sec, $min, $hr, $day, $mon, $yr) = @_;
+    (date2jd ($day, $mon, $yr) - JD_OF_EPOCH) * SECSPERDAY +
+    (($hr || 0) * 60 + ($min || 0)) * 60 + ($sec || 0);
+}
+
+
 =item $rad = deg2rad ($degr)
 
 This subroutine converts degrees to radians.
@@ -183,8 +286,6 @@ $sum
 
 =item $seconds = dynamical_delta ($time);
 
-=for comment help syntax-highlighting editor "
-
 This method returns the difference between dynamical and universal time
 at the given universal time. That is,
 
@@ -195,8 +296,6 @@ if $time is universal time.
 The algorithm is from Jean Meeus' "Astronomical Algorithms", 2nd
 Edition, Chapter 10, page 78.
 
-=for comment help syntax-highlighting editor "
-
 =cut
 
 sub dynamical_delta {
@@ -208,6 +307,53 @@ my $correction = .37 * ($year - 2100);	# Meeus' correction to (10.2)
 }
 
 
+=item ($sec, $min, $hr, $day, $mon, $yr, $wday, $yday, 0) = epoch2datetime ($epoch)
+
+This convenience subroutine converts the given time in seconds from the
+system epoch to the corresponding date and time. It is implemented in
+terms of jd2date (), with the year and month returned from that
+subroutine. The day is a whole number, with the fractional part
+converted to hours, minutes, and seconds. The $wday is the day of the
+week, with Sunday being 0. The $yday is the day of the year, with
+January 1 being 0. The trailing 0 is the summer time (or daylight saving
+time) indicator which is always 0 to be consistent with gmtime.
+
+If called in scalar context, it returns the date formatted by
+POSIX::strftime, using the format string in $DATETIMEFORMAT.
+
+The functionality is similar to gmtime, but lacks gmtime's limited date
+range.
+
+The input must convert to a non-negative Julian date. The exact lower
+limit depends on the system, but is computed by -(JD_OF_EPOCH * 86400).
+For Unix systems with an epoch of January 1 1970, this is -210866760000.
+
+Additional algorithms for day of week and day of year come from Jean
+Meeus' "Astronomical Algorithms", 2nd Edition, Chapter 7 (Julian Day),
+page 65.
+
+=cut
+
+sub epoch2datetime {
+    my $day = floor ($_[0] / SECSPERDAY);
+    my $sec = $_[0] - $day * SECSPERDAY;
+    ($day, my $mon, my $yr, my $greg, my $leap) = jd2date (
+	my $jd = $day + JD_OF_EPOCH);
+    $day = floor ($day + .5);
+    my $min = floor ($sec / 60);
+    $sec = $sec - $min * 60;
+    my $hr = floor ($min / 60);
+    $min = $min - $hr * 60;
+    my $wday = ($jd + 1.5) % 7;
+    my $yd = floor (275 * ($mon + 1) / 9) - (2 - $leap) * floor (($mon +
+	10) / 12) + $day - 31;
+    wantarray and return ($sec, $min, $hr, $day, $mon, $yr, $wday, $yd,
+	0);
+    return strftime ($DATETIMEFORMAT, $sec, $min, $hr, $day, $mon, $yr,
+	$wday, $yd, 0);
+}
+
+
 =item $seconds = equation_of_time ($time);
 
 This method returns the equation of time at the given B<dynamical>
@@ -293,9 +439,6 @@ sub find_first_true {
 }
 
 
-
-=for comment help syntax-highlighting editor "
-
 =item $difference = intensity_to_magnitude ($ratio)
 
 This method converts a ratio of light intensities to a difference in
@@ -307,8 +450,6 @@ represents a brighter star) you get back a positive result for an
 intensity ratio less than 1, and a negative result for an intensity
 ratio greater than 1.
 
-=for comment help syntax-highlighting editor "
-
 =cut
 
 {	# Begin local symbol block
@@ -319,7 +460,83 @@ ratio greater than 1.
 }
 
 
-=for comment help syntax-highlighting editor "
+=item ($day, $mon, $yr, $greg, $leap) = jd2date ($jd)
+
+This subroutine converts the given Julian day to the corresponding date.
+The returns are year - 1900, month (0 to 11), day (which may have a
+fractional part), a Gregorian calendar indicator which is true if the
+date is in the Gregorian calendar and false if it is in the Julian
+calendar, and a leap (or bissextile) year indicator which is true if the
+year is a leap year and false otherwise. The year 1 BC is returned as
+-1900 (i.e. as year 0), and so on. The date will probably have a
+fractional part (e.g. 2006 1 1.5 for noon January first 2006).
+
+If the $jd is before $JD_GREGORIAN, the date will be in the Julian
+calendar; otherwise it will be in the Gregorian calendar.
+
+The input may not be less than 0.
+
+The algorithm is from Jean Meeus' "Astronomical Algorithms", second
+edition, chapter 7 ("Julian Day"), pages 63ff, but the month is
+zero-based, not 1-based, and the year is 1900-based.
+
+=cut
+
+sub jd2date {
+    my $mod_jd = $_[0] + 0.5;
+    my $Z = floor ($mod_jd);
+    my $F = $mod_jd - $Z;
+    my $A = (my $julian = $Z < $JD_GREGORIAN) ? $Z : do {
+	my $alpha = floor (($Z - 1867216.25)/36524.25);
+	$Z + 1 + $alpha - floor ($alpha / 4);
+    };
+    my $B = $A + 1524;
+    my $C = floor (($B - 122.1) / 365.25);
+    my $D = floor (365.25 * $C);
+    my $E = floor (($B - $D) / 30.6001);
+    my $day = $B - $D - floor (30.6001 * $E) + $F;
+    my $mon = $E < 14 ? $E - 1 : $E - 13;
+    my $yr = $mon > 2 ? $C - 4716 : $C - 4715;
+    ($day, $mon - 1, $yr - 1900, !$julian, ($julian ? !($yr % 4) : (
+		$yr % 400 ? $yr % 100 ? !($yr % 4) : 0 : 1)));
+##	% 400 ? 1 : $yr % 100 ? 0 : !($yr % 4))));
+}
+
+
+=item ($sec, $min, $hr, $day, $mon, $yr, $wday, $yday, 0) = jd2datetime ($jd)
+
+This convenience subroutine converts the given Julian day to the
+corresponding date and time. All this really does is converts its
+argument to seconds since the system epoch, and pass off to
+epoch2datetime().
+
+The input may not be less than 0.
+
+=cut
+
+sub jd2datetime {
+
+=begin comment
+
+    my ($day, $mon, $yr) = jd2date (@_);
+    my $hr = $day;
+    $day = floor ($day);
+    my $min = $hr = ($hr - $day) * 24;
+    $hr = floor ($hr);
+    my $sec = $min = ($min - $hr) * 60;
+    $min = floor ($min);
+    $sec = ($sec - $min) * 60;
+    wantarray and return ($sec, $min, $hr, $day, $mon, $yr);
+    return strftime ($DATETIMEFORMAT, $yr, $mon, $day, $hr, $min, $sec);
+
+=end comment
+
+=cut
+
+    $_[0] = ($_[0] - JD_OF_EPOCH) * SECSPERDAY;
+    goto &epoch2datetime;
+}
+
 
 =item $century = jcent2000 ($time);
 
@@ -328,9 +545,6 @@ are expressed in terms of the number of Julian centuries from epoch
 J2000.0 (e.g equations 12.1, 22.1). This subroutine encapsulates
 that calculation.
 
-
-=for comment help syntax-highlighting editor "
-
 =cut
 
 sub jcent2000 {
@@ -355,8 +569,6 @@ sub jday2000 {
 }
 
 
-=for comment help syntax-highlighting editor "
-
 =item $jd = julianday ($time);
 
 This subroutine converts a Perl date to a Julian day number.
@@ -364,14 +576,34 @@ This subroutine converts a Perl date to a Julian day number.
 The computation makes use of information from Jean Meeus' "Astronomical
 Algorithms", 2nd Edition, Chapter 7, page 62.
 
-=for comment help syntax-highlighting editor "
-
 =cut
 
 sub julianday {
 jday2000($_[0]) + 2_451_545.0	#   Meeus p. 62
 }
 
+=item $rslt = load_module ($module_name)
+
+This convenience method loads the named module (using 'require'),
+throwing an exception if the load fails. Results are cached, and
+subsequent attempts to load the same module simply give the cached
+results.
+
+=cut
+
+{	# Local symbol block. Oh, for 5.10 and state variables.
+    my %error;
+    my %rslt;
+    sub load_module {
+	my  ($module) = @_;
+	exists $error{$module} and croak $error{$module};
+	exists $rslt{$module} and return $rslt{$module};
+	$rslt{$module} = eval "require $module";
+	$@ and croak ($error{$module} = $@);
+	return $rslt{$module};
+    }
+}	# End local symbol block.
+
 
 =item $theta = mod2pi ($theta)
 
@@ -385,8 +617,6 @@ $_[0] - floor ($_[0] / TWOPI) * TWOPI;
 }
 
 
-=for comment help syntax-highlighting editor "
-
 =item $delta_psi = nutation_in_longitude ($time)
 
 This subroutine calculates the nutation in longitude (delta psi) for
@@ -396,8 +626,6 @@ The algorithm comes from Jean Meeus' "Astronomical Algorithms", 2nd
 Edition, Chapter 22, pages 143ff. Meeus states that it is good to
 0.5 seconds of arc.
 
-=for comment help syntax-highlighting editor "
-
 =cut
 
 sub nutation_in_longitude {
@@ -417,8 +645,6 @@ $delta_psi;
 }
 
 
-=for comment help syntax-highlighting editor "
-
 =item $delta_epsilon = nutation_in_obliquity ($time)
 
 This subroutine calculates the nutation in obliquity (delta epsilon)
@@ -428,8 +654,6 @@ The algorithm comes from Jean Meeus' "Astronomical Algorithms", 2nd
 Edition, Chapter 22, pages 143ff. Meeus states that it is good to
 0.1 seconds of arc.
 
-=for comment help syntax-highlighting editor "
-
 =cut
 
 sub nutation_in_obliquity {
@@ -449,8 +673,6 @@ $delta_epsilon;
 }
 
 
-=for comment help syntax-highlighting editor "
-
 =item $epsilon = obliquity ($time)
 
 This subroutine calculates the obliquity of the ecliptic in radians at
@@ -460,8 +682,6 @@ The algorithm comes from Jean Meeus' "Astronomical Algorithms", 2nd
 Edition, Chapter 22, pages 143ff. The conversion from universal to
 dynamical time comes from chapter 10, equation 10.2  on page 78.
 
-=for comment help syntax-highlighting editor "
-
 =cut
 
 use constant E0BASE => (21.446 / 60 + 26) / 60 + 23;
@@ -479,7 +699,6 @@ my $epsilon0 = deg2rad (((0.001813 * $T - 0.00059) * $T - 46.8150)
 $epsilon0 + $delta_epsilon;
 }
 
-=for comment help syntax-highlighting editor "
 
 =item $radians = omega ($time);
 
@@ -489,8 +708,6 @@ of the Moon's mean orbit at the given B<dynamical> time.
 The algorithm comes from Jean Meeus' "Astronomical Algorithms", 2nd
 Edition, Chapter 22, pages 143ff.
 
-=for comment help syntax-highlighting editor "
-
 =cut
 
 sub omega {
@@ -550,8 +767,6 @@ thetag (timegm (0, 0, 0, (gmtime $_[0])[3 .. 5]));
 }
 
 
-=for comment help syntax-highlighting editor "
-
 =item $value = thetag ($time);
 
 This subroutine returns the Greenwich hour angle of the mean equinox at
@@ -560,8 +775,6 @@ the given time.
 The algorithm comes from Jean Meeus' "Astronomical Algorithms", 2nd
 Edition, equation 12.4, page 88.
 
-=for comment help syntax-highlighting editor "
-
 =cut
 
 
@@ -601,6 +814,8 @@ Thomas R. Wyant, III (F<wyant at cpan dot org>)
 Copyright 2005, 2006, 2007 by Thomas R. Wyant, III
 (F<wyant at cpan dot org>). All rights reserved.
 
+=head1 LICENSE
+
 This module is free software; you can use it, redistribute it and/or
 modify it under the same terms as Perl itself. Please see
 L<http://perldoc.perl.org/index-licence.html> for the current licenses.
